/* *********************************************************************************************************************
	  ____         ____
	 /    \____/  /    \          Written by: .. Jay Brockway
	 \/     |      |    \         Date: ........ 2021/01/31
			|      |____/         Revised: ..... 2021/07/24
			|      |     \
			|      |      \
	   ____ |      |_     /
	  /    \| #   /  \___/ #
********************************************************************************************************************* */
#include <cstddef>					// for NULL
#include <Arduino.h>				// for millis()
#include "JB_Publisher_Subscriber.h"


namespace JB
{
	/* ****************************************************************************************************************
	Subscriber Class Definitions
	***************************************************************************************************************** */

	/* == Class Constructors ======================================================================================== */
	Subscriber::Subscriber(){}


	Subscriber::Subscriber(void (*subFunc)())
	{
		this->subFunc = subFunc;
	}
}


namespace JB
{
	/* *****************************************************************************************************************
	Publisher Class Definitions
	***************************************************************************************************************** */

	/* == Class Constructors ======================================================================================== */
	Publisher::Publisher() {}


    /* == Class Functions =========================================================================================== */
	/* -------------------------------------------------------------------------------------------------------------- */
	void Publisher::Add(Subscriber* subscriber)
	{
		if (NULL != subscriber)
		{
			subscriber->next = NULL;        // Init to NULL.

			if (NULL == this->head)         // Head is set to point at the first subscriber.
			{
				this->head = subscriber;
			}

			if (NULL != this->tail)
			{
				// The current tail is the previously added link.
				// Point the previous link to the new link.
				this->tail->next = subscriber;
			}

			this->tail = subscriber;        // Update tail to point at the latest subscriber.
		}
	}


	/* -------------------------------------------------------------------------------------------------------------- */
	void Publisher::Remove(Subscriber* subscriber)
	{
		// If linked list has nodes.
		if ((NULL != subscriber) && (NULL != this->head))
		{
			if (subscriber == this->head)
			{
				this->head = NULL;

				// If subscriber is the head and tail it is then the only node.
				if (subscriber == this->tail)
				{
					this->tail = NULL;
				}
			}
			else
			{
				// Subscriber is not the first node so current starts with the second node.
				Subscriber* previous = this->head;
				Subscriber* current = this->head->next;

				do
				{
					if (subscriber == current)
					{
						previous->next = current->next;

						// If subscriber is the current tail then set tail to previous.
						if (subscriber == this->tail)
						{
							this->tail = previous;
						}
					}

					previous = current;
					current = current->next;
				} while (NULL != current);
			}
		}
	}


	/* -------------------------------------------------------------------------------------------------------------- */
	void Publisher::Clear()
	{
		this->head = NULL;
		this->tail = NULL;
	}


	/* -------------------------------------------------------------------------------------------------------------- */
	void Publisher::Notify()
	{
		if (NULL != this->head)
		{
			Subscriber* current = this->head;
			do
			{
				current->subFunc();
				current = current->next;
			} while (NULL != current);
		}
	}
}
