/* *********************************************************************************************************************
	  ____         ____
	 /    \____/  /    \          Written by: .. Jay Brockway
	 \/     |      |    \         Date: ........ 2021/01/31
			|      |____/         Revised: ..... 2021/07/24
			|      |     \
			|      |      \
	   ____ |      |_     /
	  /    \| #   /  \___/ #

Task Scheduler
********************************************************************************************************************* */
#include <cstddef>					// for NULL
#include <Arduino.h>				// for millis()
#include "JB_TaskScheduler.h"


namespace JB
{
	/* ** **************************************************************************************************************
	Task Class Definitions
	***************************************************************************************************************** */

	/* == Class Constructors ======================================================================================== */
	Task::Task(){}


	Task::Task(void (*taskFunc)(), int period)
	{
		this->taskFunc = taskFunc;
		this->period = period;
		this->nextTime = millis();
	}


    /* == Class Functions =========================================================================================== */
	/* -------------------------------------------------------------------------------------------------------------- */
	void Task::Schedule()
	{
		this->nextTime = millis() + this->period;
	}
}


namespace JB
{
	/* *****************************************************************************************************************
	Scheduler Class Definitions
	***************************************************************************************************************** */
    
	/* == Class Constructors ======================================================================================== */
	Scheduler::Scheduler(){}


	/* == Class Functions =========================================================================================== */
	/* -------------------------------------------------------------------------------------------------------------- */
	void Scheduler::Add(Task* task)
	{
		if (NULL != task)
		{
			task->next = NULL;              // Init to NULL.

			if (NULL == this->head)         // Head is set to point at the first task.
			{
				this->head = task;
			}

			if (NULL != this->tail)
			{
				// The current tail is the previously added link.
				// Point the previous link to the new link.
				this->tail->next = task;
			}

			this->tail = task;              // Update tail to point at the last task.
		}
	}


	/* -------------------------------------------------------------------------------------------------------------- */
	void Scheduler::Remove(Task* task)
	{
		// If linked list has nodes.
		if ((NULL != task) && (NULL != this->head))
		{
			if (task == this->head)
			{
				this->head = NULL;

				// If task is the head and tail it is then the only node.
				if (task == this->tail)
				{
					this->tail = NULL;
				}
			}
			else
			{
				// Task is not the first node so current starts with the second node.
				Task* previous = this->head;
				Task* current = this->head->next;

				do
				{
					if (task == current)
					{
						previous->next = current->next;

						// If task is the current tail then set tail to previous.
						if (task == this->tail)
						{
							this->tail = previous;
						}
					}

					previous = current;
					current = current->next;
				} while (NULL != current);
			}
		}
	}


	/* -------------------------------------------------------------------------------------------------------------- */
	void Scheduler::Run()
	{
		if (NULL != this->head)
		{
			Task* taskCurrent = this->head;
			do
			{
				// Only call function if scheduled time has past.
				if (millis() > taskCurrent->nextTime)
				{
					taskCurrent->Schedule();
					taskCurrent->taskFunc();
				}
				taskCurrent = taskCurrent->next;
			} while (NULL != taskCurrent);
		}
	}
}